"""Tests for Ralph cli."""
import json
import logging
from pathlib import Path

import pytest
from click.exceptions import BadParameter
from click.testing import CliRunner
from elasticsearch.helpers import bulk, scan
from pydantic import ValidationError

from ralph.backends.storage.fs import FSStorage
from ralph.backends.storage.ldp import LDPStorage
from ralph.cli import (
    CommaSeparatedKeyValueParamType,
    CommaSeparatedTupleParamType,
    JSONStringParamType,
    cli,
)
from ralph.conf import settings
from ralph.exceptions import ConfigurationException
from ralph.models.edx.navigational.statements import UIPageClose
from ralph.models.xapi.navigation.statements import PageTerminated

from tests.fixtures.backends import (
    ES_TEST_HOSTS,
    ES_TEST_INDEX,
    WS_TEST_HOST,
    WS_TEST_PORT,
)
from tests.fixtures.hypothesis_strategies import custom_given

test_logger = logging.getLogger("ralph")


def test_cli_comma_separated_key_value_param_type():
    """Tests the CommaSeparatedKeyValueParamType custom parameter type."""
    param_type = CommaSeparatedKeyValueParamType()

    # Bad options
    with pytest.raises(
        BadParameter,
        match=(
            "You should provide key=value pairs separated by commas, "
            "e.g. foo=bar,bar=2"
        ),
    ):
        param_type.convert("foo=bar,baz", None, None)
    with pytest.raises(
        BadParameter,
        match=(
            "You should provide key=value pairs separated by commas, "
            "e.g. foo=bar,bar=2"
        ),
    ):
        param_type.convert("foo=bar,", None, None)

    # Option already a dictionary (from defaults)
    assert param_type.convert({"foo": "bar"}, None, None) == {"foo": "bar"}

    # Simple string parsing
    assert param_type.convert("foo=bar", None, None) == {
        "foo": "bar",
    }
    assert param_type.convert("foo=bar,baz=spam", None, None) == {
        "foo": "bar",
        "baz": "spam",
    }
    assert param_type.convert("foo=bar,baz=", None, None) == {
        "foo": "bar",
        "baz": None,
    }

    # Boolean type casting
    assert param_type.convert("foo=true,bar=false", None, None) == {
        "foo": True,
        "bar": False,
    }
    assert param_type.convert("foo=True,bar=False", None, None) == {
        "foo": True,
        "bar": False,
    }

    # Integer type casting
    assert param_type.convert("foo=42,bar=1", None, None) == {
        "foo": 42,
        "bar": 1,
    }

    # Float type casting
    assert param_type.convert("foo=4.2,bar=12.3", None, None) == {
        "foo": 4.2,
        "bar": 12.3,
    }

    # Mixed types casting
    assert param_type.convert("foo=lol,bar=12.3,baz=1,spam=True", None, None) == {
        "foo": "lol",
        "bar": 12.3,
        "baz": 1,
        "spam": True,
    }


@pytest.mark.parametrize(
    "value,expected",
    [
        ("foo", ("foo",)),
        ("foo,bar", ("foo", "bar")),
        ("foo,bar,baz", ("foo", "bar", "baz")),
        (("foo", "bar", "baz"), ("foo", "bar", "baz")),
    ],
)
def test_cli_comma_separated_tuple_param_type_with_valid_input(value, expected):
    """Tests the CommaSeparatedTupleParamType parameter type with valid input."""
    param_type = CommaSeparatedTupleParamType()
    assert param_type.convert(value, None, None) == expected


@pytest.mark.parametrize("value", [None, {}, ["foo"], 10, True])
def test_cli_comma_separated_tuple_param_type_with_invalid_input(value):
    """Tests the CommaSeparatedTupleParamType parameter type with invalid input."""
    param_type = CommaSeparatedTupleParamType()
    with pytest.raises(
        BadParameter,
        match="You should provide values separated by commas, e.g. foo,bar,baz",
    ):
        param_type.convert(value, None, None)


@pytest.mark.parametrize("value,expected", [('{"foo": "bar"}', {"foo": "bar"})])
def test_cli_json_string_param_type_with_valid_input(value, expected):
    """Tests the JSONStringParamType custom parameter type with valid input."""
    param_type = JSONStringParamType()
    assert param_type.convert(value, None, None) == expected


@pytest.mark.parametrize("value", ["foo", None, {}])
def test_cli_json_string_param_type_with_invalid_input(value):
    """Tests the JSONStringParamType custom parameter type with invalid input."""
    param_type = JSONStringParamType()
    with pytest.raises(
        BadParameter,
        match="You should provide a valid JSON string as input",
    ):
        param_type.convert(value, None, None)


def test_cli_help_option():
    """Tests ralph --help command."""
    runner = CliRunner()
    result = runner.invoke(cli, ["--help"])

    assert result.exit_code == 0
    assert (
        (
            "-v, --verbosity LVL  Either CRITICAL, ERROR, WARNING, INFO (default) or "
            "DEBUG"
        )
    ) in result.output


def test_cli_auth_command_without_writing_auth_file():
    """Test ralph auth command when credentials are displayed in the tty."""
    runner = CliRunner()
    result = runner.invoke(cli, ["auth", "-u", "foo", "-p", "bar", "-s", "foo_scope"])

    assert result.exit_code == 0

    credentials = json.loads("".join(result.output.split("\n")[-8:-1]))
    assert credentials["username"] == "foo"
    assert "hash" in credentials
    assert len(credentials["scopes"]) == 1
    assert credentials["scopes"][0] == "foo_scope"

    # Test multiple scopes
    result = runner.invoke(
        cli, ["auth", "-u", "foo", "-p", "bar", "-s", "foo_scope", "-s", "admin_scope"]
    )

    assert result.exit_code == 0

    credentials = json.loads("".join(result.output.split("\n")[-9:-1]))
    assert credentials["username"] == "foo"
    assert "hash" in credentials
    assert len(credentials["scopes"]) == 2
    assert credentials["scopes"][0] == "foo_scope"
    assert credentials["scopes"][1] == "admin_scope"


# pylint: disable=invalid-name,unused-argument
def test_cli_auth_command_when_writing_auth_file(fs):
    """Test ralph auth command when credentials are written in the authentication
    file.
    """
    runner = CliRunner()

    # The authentication file does not exist
    assert Path(settings.AUTH_FILE).exists() is False
    result = runner.invoke(
        cli, ["auth", "-u", "foo", "-p", "bar", "-s", "foo_scope", "-w"]
    )
    assert result.exit_code == 0
    assert Path(settings.AUTH_FILE).exists() is True
    with Path(settings.AUTH_FILE).open(encoding="utf-8") as auth_file:
        all_credentials = json.loads("\n".join(auth_file.readlines()))
    assert len(all_credentials) == 1
    credentials = all_credentials[0]
    assert credentials["username"] == "foo"
    assert "hash" in credentials
    assert len(credentials["scopes"]) == 1
    assert credentials["scopes"][0] == "foo_scope"

    # Add a new user
    result = runner.invoke(
        cli,
        [
            "auth",
            "-u",
            "lol",
            "-p",
            "baz",
            "-s",
            "lol_scope",
            "-s",
            "admin_scope",
            "-w",
        ],
    )
    assert result.exit_code == 0
    with Path(settings.AUTH_FILE).open(encoding="utf-8") as auth_file:
        all_credentials = json.loads("\n".join(auth_file.readlines()))
    assert len(all_credentials) == 2
    credentials = all_credentials[0]
    assert credentials["username"] == "foo"
    assert "hash" in credentials
    assert len(credentials["scopes"]) == 1
    assert credentials["scopes"][0] == "foo_scope"
    credentials = all_credentials[1]
    assert credentials["username"] == "lol"
    assert "hash" in credentials
    assert len(credentials["scopes"]) == 2
    assert credentials["scopes"][0] == "lol_scope"
    assert credentials["scopes"][1] == "admin_scope"


# pylint: disable=invalid-name
def test_cli_auth_command_when_writing_auth_file_whith_incorrect_auth_file(fs):
    """Test ralph auth command when credentials are written in the authentication
    file with a badly formatted original authentication file.
    """
    runner = CliRunner()

    # Create the base auth file path in the fake fs with badly formatted JSON
    # file content
    contents = (
        "{\n"
        ' "username": "johndoe",\n'
        ' "hash": "secretlynothashed",\n'
        ' "scope": ["johndoe_scope"]\n'
    )
    fs.create_file(settings.AUTH_FILE, contents=contents)

    result = runner.invoke(
        cli, ["auth", "-u", "foo", "-p", "bar", "-s", "foo_scope", "-w"]
    )
    assert result.exit_code > 0
    auth_file = Path(settings.AUTH_FILE)
    assert auth_file.exists() is True
    assert auth_file.read_text(encoding="utf-8") == contents


def test_cli_extract_command_with_gelf_parser(gelf_logger):
    """Tests the extract command using the GELF parser."""
    gelf_logger.info('{"username": "foo"}')

    runner = CliRunner()
    with Path(gelf_logger.handlers[0].stream.name).open(
        encoding=settings.LOCALE_ENCODING
    ) as log_file:
        gelf_content = log_file.read()
        result = runner.invoke(cli, ["extract", "-p", "gelf"], input=gelf_content)
        assert '{"username": "foo"}' in result.output


def test_cli_extract_command_with_es_parser():
    """Tests the extract command using the ElasticSearchParser."""
    es_output = (
        "\n".join(
            [
                json.dumps(
                    {
                        "_index": ES_TEST_INDEX,
                        "_id": str(idx),
                        "_score": None,
                        "_source": {"id": idx},
                        "sort": [idx],
                    }
                )
                for idx in range(10)
            ]
        )
        + "\n"
    )

    runner = CliRunner()
    result = runner.invoke(cli, "extract -p es".split(), input=es_output)
    assert result.exit_code == 0
    assert "\n".join([json.dumps({"id": idx}) for idx in range(10)]) in result.output


@custom_given(UIPageClose)
def test_cli_validate_command_with_edx_format(event):
    """Tests the validate command using the edx format."""
    event_str = event.json()
    runner = CliRunner()
    result = runner.invoke(cli, ["validate", "-f", "edx"], input=event_str)
    assert event_str in result.output


@custom_given(UIPageClose)
@pytest.mark.parametrize("valid_uuid", ["ee241f8b-174f-5bdb-bae9-c09de5fe017f"])
def test_cli_convert_command_from_edx_to_xapi_format(valid_uuid, event):
    """Tests the convert command from edx to xapi format."""
    event_str = event.json()
    runner = CliRunner()
    command = f"-v ERROR convert -f edx -t xapi -u {valid_uuid} -p https://fun-mooc.fr"
    result = runner.invoke(cli, command.split(), input=event_str)
    assert result.exit_code == 0
    try:
        PageTerminated(**json.loads(result.output))
    except ValidationError as err:
        pytest.fail(f"Converted event is invalid: {err}")


@pytest.mark.parametrize("invalid_uuid", ["", None, 1, {}])
def test_cli_convert_command_with_invalid_uuid(invalid_uuid):
    """Tests that the convert command raises an exception when the uuid namespace is
    invalid.
    """
    runner = CliRunner()
    command = f"convert -f edx -t xapi -u '{invalid_uuid}' -p https://fun-mooc.fr"
    result = runner.invoke(cli, command.split())
    assert result.exit_code > 0
    assert isinstance(result.exception, ConfigurationException)
    assert str(result.exception) == "Invalid UUID namespace"


@cli.command()
def dummy_verbosity_check():
    """Adding a dummy command to the cli with all logging levels."""
    test_logger.critical("CRITICAL")
    test_logger.error("ERROR")
    test_logger.warning("WARNING")
    test_logger.info("INFO")
    test_logger.debug("DEBUG")


@pytest.mark.parametrize("verbosity", ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"])
def test_cli_verbosity_option_should_impact_logging_behaviour(verbosity):
    """Tests that the verbosity option impacts logging output."""
    runner = CliRunner()
    result = runner.invoke(cli, ["-v", verbosity, "dummy-verbosity-check"])
    assert verbosity in result.output


def test_cli_fetch_command_with_ldp_backend(monkeypatch):
    """Tests the fetch command using the LDP backend."""
    archive_content = {"foo": "bar"}

    def mock_read(this, name, chunk_size=500):
        """Always return the same archive."""
        # pylint: disable=unused-argument

        yield bytes(json.dumps(archive_content), encoding="utf-8")

    monkeypatch.setattr(LDPStorage, "read", mock_read)

    runner = CliRunner()
    command = "fetch -b ldp --ldp-endpoint ovh-eu a547d9b3-6f2f-4913-a872-cf4efe699a66"
    result = runner.invoke(cli, command.split())

    assert result.exit_code == 0
    assert '{"foo": "bar"}' in result.output


# pylint: disable=invalid-name
# pylint: disable=unused-argument
def test_cli_fetch_command_with_fs_backend(fs, monkeypatch):
    """Tests the fetch command using the FS backend."""
    archive_content = {"foo": "bar"}

    def mock_read(this, name, chunk_size):
        """Always return the same archive."""
        # pylint: disable=unused-argument

        yield bytes(json.dumps(archive_content), encoding="utf-8")

    monkeypatch.setattr(FSStorage, "read", mock_read)

    runner = CliRunner()
    result = runner.invoke(cli, "fetch -b fs foo".split())

    assert result.exit_code == 0
    assert '{"foo": "bar"}' in result.output


def test_cli_fetch_command_with_es_backend(es):
    """Tests ralph fetch command using the es backend."""
    # pylint: disable=invalid-name

    # Insert documents
    bulk(
        es,
        (
            {"_index": ES_TEST_INDEX, "_id": idx, "_source": {"id": idx}}
            for idx in range(10)
        ),
    )
    # As we bulk insert documents, the index needs to be refreshed before making
    # queries.
    es.indices.refresh(index=ES_TEST_INDEX)

    runner = CliRunner()
    es_hosts = ",".join(ES_TEST_HOSTS)
    es_client_options = "verify_certs=True"
    command = f"""-v ERROR fetch -b es --es-hosts {es_hosts} --es-index {ES_TEST_INDEX}
        --es-client-options {es_client_options}"""
    result = runner.invoke(cli, command.split())
    assert result.exit_code == 0
    expected = (
        "\n".join(
            [
                json.dumps(
                    {
                        "_index": ES_TEST_INDEX,
                        "_id": str(idx),
                        "_score": None,
                        "_source": {"id": idx},
                        "sort": [idx],
                    }
                )
                for idx in range(10)
            ]
        )
        + "\n"
    )

    assert expected == result.output


def test_cli_fetch_command_client_options_with_es_backend(es):
    """Tests ralph fetch command with multiple client options using the es backend."""
    # pylint: disable=invalid-name

    runner = CliRunner()
    es_client_options = "ca_certs=/path/,verify_certs=True"
    command = f"""-v ERROR fetch -b es --es-client-options {es_client_options}"""
    result = runner.invoke(cli, command.split())
    assert result.exit_code == 1
    assert "TLS options require scheme to be 'https'" in str(result.exception)


def test_cli_fetch_command_with_es_backend_query(es):
    """Tests ralph fetch command using the es backend and a query."""
    # pylint: disable=invalid-name

    # Insert documents
    bulk(
        es,
        (
            {
                "_index": ES_TEST_INDEX,
                "_id": idx,
                "_source": {"id": idx, "modulo": idx % 2},
            }
            for idx in range(10)
        ),
    )
    # As we bulk insert documents, the index needs to be refreshed before making
    # queries.
    es.indices.refresh(index=ES_TEST_INDEX)

    runner = CliRunner()
    es_hosts = ",".join(ES_TEST_HOSTS)
    query = {"query": {"query": {"term": {"modulo": 0}}}}
    query_str = json.dumps(query, separators=(",", ":"))

    command = (
        "-v ERROR "
        "fetch "
        "-b es "
        f"--es-hosts {es_hosts} "
        f"--es-index {ES_TEST_INDEX} "
        f"--query {query_str}"
    )
    result = runner.invoke(cli, command.split())
    assert result.exit_code == 0
    expected = (
        "\n".join(
            [
                json.dumps(
                    {
                        "_index": ES_TEST_INDEX,
                        "_id": str(idx),
                        "_score": None,
                        "_source": {"id": idx, "modulo": 0},
                        "sort": [idx],
                    }
                )
                for idx in range(0, 10, 2)
            ]
        )
        + "\n"
    )

    assert expected == result.output


def test_cli_fetch_command_with_ws_backend(events, ws):
    """Tests ralph fetch command using the ws backend."""
    runner = CliRunner()
    result = runner.invoke(
        cli,
        ["fetch", "-b", "ws", "--ws-uri", f"ws://{WS_TEST_HOST}:{WS_TEST_PORT}"],
    )
    assert "\n".join([json.dumps(event) for event in events]) in result.output


def test_cli_list_command_with_ldp_backend(monkeypatch):
    """Tests the list command using the LDP backend."""
    archive_list = [
        "5d5c4c93-04a4-42c5-9860-f51fa4044aa1",
        "997db3eb-b9ca-485d-810f-b530a6cef7c6",
    ]
    archive_list_details = [
        {
            "archiveId": "5d5c4c93-04a4-42c5-9860-f51fa4044aa1",
            "createdAt": "2020-06-18T04:38:59.436634+02:00",
            "filename": "2020-06-16.gz",
            "md5": "01585b394be0495e38dbb60b20cb40a9",
            "retrievalDelay": 0,
            "retrievalState": "sealed",
            "sha256": "645d8e21e6fdb8aa7ffc5c[...]9ce612d06df8dcf67cb29a45ca",
            "size": 67906662,
        },
        {
            "archiveId": "997db3eb-b9ca-485d-810f-b530a6cef7c6",
            "createdAt": "2020-06-18T04:38:59.436634+02:00",
            "filename": "2020-06-17.gz",
            "md5": "01585b394be0495e38dbb60b20cb40a9",
            "retrievalDelay": 0,
            "retrievalState": "sealed",
            "sha256": "645d8e21e6fdb8aa7ffc5c[...]9ce612d06df8dcf67cb29a45ca",
            "size": 67906662,
        },
    ]

    def mock_list(this, details=False, new=False):
        """Mock LDP backend list method."""
        # pylint: disable=unused-argument

        response = archive_list
        if details:
            response = archive_list_details
        if new:
            response = response[1:]
        return response

    monkeypatch.setattr(LDPStorage, "list", mock_list)

    runner = CliRunner()

    # List archives with default options
    result = runner.invoke(cli, ["list", "-b", "ldp", "--ldp-endpoint", "ovh-eu"])
    assert result.exit_code == 0
    assert "\n".join(archive_list) in result.output

    # List archives with detailed output
    result = runner.invoke(cli, ["list", "-b", "ldp", "--ldp-endpoint", "ovh-eu", "-D"])
    assert result.exit_code == 0
    assert (
        "\n".join(json.dumps(detail) for detail in archive_list_details)
        in result.output
    )

    # List new archives only
    result = runner.invoke(cli, ["list", "-b", "ldp", "--ldp-endpoint", "ovh-eu", "-n"])
    assert result.exit_code == 0
    assert "997db3eb-b9ca-485d-810f-b530a6cef7c6" in result.output
    assert "5d5c4c93-04a4-42c5-9860-f51fa4044aa1" not in result.output

    # Edge case: stream contains no archive
    monkeypatch.setattr(LDPStorage, "list", lambda this, details, new: ())
    result = runner.invoke(cli, ["list", "-b", "ldp", "--ldp-endpoint", "ovh-eu"])
    assert result.exit_code == 0
    assert "Configured ldp backend contains no archive" in result.output


# pylint: disable=invalid-name
# pylint: disable=unused-argument
def test_cli_list_command_with_fs_backend(fs, monkeypatch):
    """Tests the list command using the LDP backend."""
    archive_list = [
        "file1",
        "file2",
    ]
    archive_list_details = [
        {
            "filename": "file1",
            "modified_at": "1604661859",
            "size": 350,
        },
        {
            "filename": "file2",
            "created_at": "1604935848",
            "size": 25000,
        },
    ]

    def mock_list(this, details=False, new=False):
        """Mock LDP backend list method."""
        # pylint: disable=unused-argument

        response = archive_list
        if details:
            response = archive_list_details
        if new:
            response = response[1:]
        return response

    monkeypatch.setattr(FSStorage, "list", mock_list)

    runner = CliRunner()

    # List archives with default options
    result = runner.invoke(cli, ["list", "-b", "fs"])
    assert result.exit_code == 0
    assert "\n".join(archive_list) in result.output

    # List archives with detailed output
    result = runner.invoke(cli, ["list", "-b", "fs", "-D"])
    assert result.exit_code == 0
    assert (
        "\n".join(json.dumps(detail) for detail in archive_list_details)
        in result.output
    )

    # List new archives only
    result = runner.invoke(cli, ["list", "-b", "fs", "-n"])
    assert result.exit_code == 0
    assert "file2" in result.output
    assert "file1" not in result.output

    # Edge case: stream contains no archive
    monkeypatch.setattr(FSStorage, "list", lambda this, details, new: ())
    result = runner.invoke(cli, ["list", "-b", "fs"])
    assert result.exit_code == 0
    assert "Configured fs backend contains no archive" in result.output


# pylint: disable=invalid-name
def test_cli_push_command_with_fs_backend(fs):
    """Tests the push command using the FS backend."""
    fs.create_dir(str(settings.APP_DIR))

    filename = Path("file1")
    file_path = Path(settings.BACKENDS.STORAGE.FS.PATH) / filename

    # Create a file
    runner = CliRunner()
    result = runner.invoke(cli, "push -b fs file1".split(), input="test content")

    assert result.exit_code == 0

    with file_path.open("r", encoding=settings.LOCALE_ENCODING) as test_file:
        content = test_file.read()

    assert "test content" in content

    # Trying to create the same file without -f should raise an error
    runner = CliRunner()
    result = runner.invoke(cli, "push -b fs file1".split(), input="other content")
    assert result.exit_code == 1
    assert "file1 already exists and overwrite is not allowed" in result.output

    # Try to create the same file with -f
    runner = CliRunner()
    result = runner.invoke(cli, "push -b fs -f file1".split(), input="other content")

    assert result.exit_code == 0

    with file_path.open("r", encoding=settings.LOCALE_ENCODING) as test_file:
        content = test_file.read()

    assert "other content" in content


def test_cli_push_command_with_es_backend(es):
    """Tests ralph push command using the es backend."""
    # pylint: disable=invalid-name

    # Documents
    records = [{"id": idx} for idx in range(10)]

    runner = CliRunner()
    es_hosts = ",".join(ES_TEST_HOSTS)
    result = runner.invoke(
        cli,
        f"push -b es --es-hosts {es_hosts} --es-index {ES_TEST_INDEX}".split(),
        input="\n".join(json.dumps(record) for record in records),
    )
    assert result.exit_code == 0

    # As we bulk insert documents, the index needs to be refreshed before making
    # queries.
    es.indices.refresh(index=ES_TEST_INDEX)
    documents = list(scan(es, index=ES_TEST_INDEX, size=10))

    assert len(documents) == 10
    assert [document.get("_source") for document in documents] == records


@pytest.mark.parametrize("host_,port_", [("0.0.0.0", "8000"), ("127.0.0.1", "80")])
def test_cli_runserver_command_with_host_and_port_arguments(host_, port_, monkeypatch):
    """Tests the ralph runserver command should consider the host and port arguments."""

    def mock_uvicorn_run(_, host=None, port=None, **kwargs):
        """Mocks uvicorn.run asserting host and port values."""
        assert host == host_
        assert port == int(port_)

    monkeypatch.setattr("ralph.cli.uvicorn.run", mock_uvicorn_run)

    runner = CliRunner()
    result = runner.invoke(cli, f"runserver -h {host_} -p {port_} -b es".split())
    assert result.exit_code == 0
    assert f"Running API server on {host_}:{port_} with es backend" in result.output


def test_cli_runserver_command_environment_file_generation(monkeypatch):
    """Tests the ralph runserver command should create the expected environment file."""

    def mock_uvicorn_run(_, env_file=None, **kwargs):
        """Mocks uvicorn.run asserting environment file content."""
        with open(env_file, mode="r", encoding=settings.LOCALE_ENCODING) as file:
            env_lines = [
                f"RALPH_RUNSERVER_BACKEND={settings.RUNSERVER_BACKEND}\n",
                "RALPH_BACKENDS__DATABASE__ES__INDEX=foo\n",
                "RALPH_BACKENDS__DATABASE__ES__CLIENT_OPTIONS__verify_certs=True\n",
                "RALPH_BACKENDS__DATABASE__CLICKHOUSE__EVENT_TABLE_NAME="
                f"{settings.BACKENDS.DATABASE.CLICKHOUSE.EVENT_TABLE_NAME}\n",
                "RALPH_BACKENDS__DATABASE__CLICKHOUSE__DATABASE="
                f"{settings.BACKENDS.DATABASE.CLICKHOUSE.DATABASE}\n",
                "RALPH_BACKENDS__DATABASE__CLICKHOUSE__PORT="
                f"{settings.BACKENDS.DATABASE.CLICKHOUSE.PORT}\n",
                "RALPH_BACKENDS__DATABASE__CLICKHOUSE__HOST="
                f"{settings.BACKENDS.DATABASE.CLICKHOUSE.HOST}\n",
                "RALPH_BACKENDS__DATABASE__MONGO__COLLECTION="
                f"{settings.BACKENDS.DATABASE.MONGO.COLLECTION}\n",
                "RALPH_BACKENDS__DATABASE__MONGO__DATABASE="
                f"{settings.BACKENDS.DATABASE.MONGO.DATABASE}\n",
                "RALPH_BACKENDS__DATABASE__MONGO__CONNECTION_URI="
                f"{settings.BACKENDS.DATABASE.MONGO.CONNECTION_URI}\n",
                "RALPH_BACKENDS__DATABASE__ES__OP_TYPE="
                f"{settings.BACKENDS.DATABASE.ES.OP_TYPE}\n",
                "RALPH_BACKENDS__DATABASE__ES__HOSTS="
                f"{','.join(settings.BACKENDS.DATABASE.ES.HOSTS)}\n",
            ]
            assert file.readlines() == env_lines

    monkeypatch.setattr("ralph.cli.uvicorn.run", mock_uvicorn_run)
    runner = CliRunner()
    result = runner.invoke(
        cli,
        "runserver -b es --es-index foo --es-client-options verify_certs=True".split(),
    )
    assert result.exit_code == 0



"""Authenticated user for the Ralph API."""

from typing import List, Optional

from pydantic import BaseModel


class AuthenticatedUser(BaseModel):
    """Pydantic model for user authentication.

    Attributes:
        username (str): Consists of the username of the current user.
        scopes (list): Consists of the scopes the user has access to.
    """

    username: str
    scopes: Optional[List[str]]



"""LRS HTTP backend for Ralph."""

import json
import logging
from itertools import chain
from typing import Iterable, Iterator, List, Optional, Union
from urllib.parse import ParseResult, parse_qs, urljoin, urlparse

import httpx
from httpx import Client, HTTPError, HTTPStatusError, RequestError
from more_itertools import chunked
from pydantic import AnyHttpUrl, BaseModel, parse_obj_as

from ralph.conf import LRSHeaders, settings
from ralph.exceptions import BackendException, BackendParameterException

from .base import (
    BaseHTTP,
    BaseQuery,
    HTTPBackendStatus,
    OperationType,
    enforce_query_checks,
)

lrs_settings = settings.BACKENDS.HTTP.LRS
logger = logging.getLogger(__name__)


class StatementResponse(BaseModel):
    """Pydantic model for `get` statements response."""

    statements: Union[List[dict], dict]
    more: Optional[str]


class LRSQuery(BaseQuery):
    """LRS body query model."""

    query: Optional[dict]


class LRSHTTP(BaseHTTP):
    """LRS HTTP backend."""

    name = "lrs"
    query = LRSQuery
    default_operation_type = OperationType.CREATE

    def __init__(  # pylint: disable=too-many-arguments
        self,
        base_url: str = lrs_settings.BASE_URL,
        username: str = lrs_settings.USERNAME,
        password: str = lrs_settings.PASSWORD,
        headers: LRSHeaders = lrs_settings.HEADERS,
        status_endpoint: str = lrs_settings.STATUS_ENDPOINT,
        statements_endpoint: str = lrs_settings.STATEMENTS_ENDPOINT,
    ):
        """Instantiates the LRS client.

        Args:
            base_url (AnyHttpUrl): LRS server URL.
            username (str): Basic auth username for LRS authentication.
            password (str): Basic auth password for LRS authentication.
            headers (dict): Headers defined for the LRS server connection.
            status_endpoint (str): Endpoint used to check server status.
            statements_endpoint (str): Default endpoint for LRS statements resource.
        """
        self.base_url = parse_obj_as(AnyHttpUrl, base_url)
        self.auth = (username, password)
        self.headers = headers
        self.status_endpoint = status_endpoint
        self.statements_endpoint = statements_endpoint

    def status(self):
        """Implements HTTP backend check for server status."""
        status_url = urljoin(self.base_url, self.status_endpoint)

        try:
            response = httpx.get(status_url)
            response.raise_for_status()
        except RequestError:
            msg = "Unable to request the server"
            logger.error(msg)
            return HTTPBackendStatus.AWAY
        except HTTPStatusError:
            msg = "Response raised an HTTP status of 4xx or 5xx"
            logger.error(msg)
            return HTTPBackendStatus.ERROR

        return HTTPBackendStatus.OK

    def list(
        self, target: str = None, details: bool = False, new: bool = False
    ) -> Iterator[Union[str, dict]]:
        # noqa: D205, D415
        """LRS HTTP backend does not support `list` method,
        calling this method will raise an error.
        """
        msg = "LRS HTTP backend does not support `list` method, cannot list from %s"
        logger.error(msg, target)
        raise NotImplementedError(msg % target)

    @enforce_query_checks
    def read(  # pylint: disable=too-many-arguments
        self,
        query: Union[str, LRSQuery] = None,
        target: str = None,
        chunk_size: Union[None, int] = 500,
        raw_output: bool = False,
        ignore_errors: bool = False,
    ) -> Iterator[Union[bytes, dict]]:
        """Get statements from LRS `target` endpoint.

        The `read` method defined in the LRS specification returns `statements` array
        and `more` IRL. The latter is required when the returned `statement` list has
        been limited.

        Args:
            query (str, LRSQuery):  The query to select records to read.
            target (str): Endpoint from which to read data (e.g. `/statements`).
                If target is `None`, `/xAPI/statements` default endpoint is used.
            chunk_size (int or None): The number of records or bytes to read in one
                batch, depending on whether the records are dictionaries or bytes.
            raw_output (bool): Controls whether to yield bytes or dictionaries.
                If the records are dictionaries and `raw_output` is set to `True`, they
                are encoded as JSON.
                If the records are bytes and `raw_output` is set to `False`, they are
                decoded as JSON by line.
            ignore_errors (bool): If `True`, errors during the read operation
                are be ignored and logged. If `False` (default), a `BackendException`
                is raised if an error occurs.
        """
        if not target:
            target = self.statements_endpoint

        query_params = {}
        if query.query:
            query_params.update(query.query)

        # Set chunk_size to limit parameter if not exist in the query
        query_params.setdefault("limit", chunk_size)

        # Create request URL
        target = ParseResult(
            scheme=urlparse(self.base_url).scheme,
            netloc=urlparse(self.base_url).netloc,
            path=target,
            query="",
            params="",
            fragment="",
        ).geturl()

        def fetch_statements(params: dict):
            while True:
                with Client(auth=self.auth, headers=self.headers) as client:
                    response = client.get(target, params=params)
                    response.raise_for_status()

                    statements_response = StatementResponse.parse_obj(response.json())
                    statements = statements_response.statements
                    statements = (
                        [statements] if not isinstance(statements, list) else statements
                    )
                    if raw_output:
                        for statement in statements:
                            yield bytes(json.dumps(statement), encoding="utf-8")
                    else:
                        for statement in statements:
                            yield statement

                    if not statements_response.more:
                        break

                    params.update(parse_qs(urlparse(statements_response.more).query))

        try:
            yield from fetch_statements(params=query_params)
        except HTTPError as error:
            msg = "Failed to fetch statements."
            logger.error("%s. %s", msg, error)
            raise BackendException(msg, *error.args) from error

    def write(  # pylint: disable=too-many-arguments
        self,
        data: Union[Iterable[bytes], Iterable[dict]],
        target: Union[None, str] = None,
        chunk_size: Union[None, int] = 500,
        ignore_errors: bool = False,
        operation_type: Union[None, OperationType] = None,
    ) -> int:
        """Writes `data` records to the `target` endpoint and returns their count.

        Args:
            data: (Iterable): The data to write.
            target (str or None): Endpoint in which to write data (e.g. `/statements`).
                If `target` is `None`, `/xAPI/statements` default endpoint is used.
            chunk_size (int or None): The number of records or bytes to write in one
                batch, depending on whether `data` contains dictionaries or bytes.
                If `chunk_size` is `None`, a default value is used instead.
            ignore_errors (bool): If `True`, errors during the write operation
                are ignored and logged. If `False` (default), a `BackendException`
                is raised if an error occurs.
            operation_type (OperationType or None): The mode of the write operation.
                If `operation_type` is `None`, the `default_operation_type` is used
                instead. See `OperationType`.
        """
        statements_count = 0

        data = iter(data)
        try:
            first_record = next(data)
        except StopIteration:
            logger.debug("Data Iterator is empty; skipping write to target.")
            return statements_count

        if not operation_type:
            operation_type = self.default_operation_type

        if operation_type in (
            OperationType.APPEND,
            OperationType.UPDATE,
            OperationType.DELETE,
        ):
            msg = f"{operation_type.value} operation type is not supported."
            logger.error(msg)
            raise BackendParameterException(msg)

        if not target:
            target = self.statements_endpoint

        target = ParseResult(
            scheme=urlparse(self.base_url).scheme,
            netloc=urlparse(self.base_url).netloc,
            path=target,
            query="",
            params="",
            fragment="",
        ).geturl()

        data = chain((first_record,), data)
        if isinstance(first_record, dict):
            data = self._parse_dict_to_bytes(data, ignore_errors)

        logger.debug(
            "Start writing to the %s endpoint (chunk size: %d)", target, chunk_size
        )

        with Client(auth=self.auth, headers=self.headers) as client:
            for chunk in chunked(data, chunk_size):
                request = client.post(target, content=chunk)
                try:
                    request.raise_for_status()
                except HTTPError as error:
                    msg = "Failed to post statements"
                    logger.error("%s. %s", msg, error)
                    raise BackendException(msg, *error.args) from error
                statements_count += len(chunk)
                logger.info("Posted %d statements", statements_count)

        return statements_count

    @staticmethod
    def _parse_dict_to_bytes(
        statements: Iterable[dict], ignore_errors: bool
    ) -> Iterator[bytes]:
        """Reads the `statements` Iterable and yields dictionaries."""
        for statement in statements:
            try:
                yield bytes(json.dumps(statement), encoding="utf-8")
            except TypeError as error:
                msg = "Failed to encode JSON: %s, for document %s"
                logger.error(msg, error, statement)
                if ignore_errors:
                    continue
                raise BackendException(msg % (error, statement)) from error



"""Base xAPI `Object` definitions (2)."""

# Nota bene: we split object definitions into `objects.py` and `unnested_objects.py`
# because of the circular dependency : objects -> context -> objects.

from datetime import datetime
from typing import List, Optional, Union

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from ..config import BaseModelWithConfig
from .agents import BaseXapiAgent
from .attachments import BaseXapiAttachment
from .contexts import BaseXapiContext
from .groups import BaseXapiGroup
from .results import BaseXapiResult
from .unnested_objects import BaseXapiUnnestedObject
from .verbs import BaseXapiVerb


class BaseXapiSubStatement(BaseModelWithConfig):
    """Pydantic model for `SubStatement` type property.

    Attributes:
        actor (dict): See BaseXapiAgent and BaseXapiGroup.
        verb (dict): See BaseXapiVerb.
        object (dict): See BaseXapiUnnestedObject.
        objecType (dict): Consists of the value `SubStatement`.
    """

    actor: Union[BaseXapiAgent, BaseXapiGroup]
    verb: BaseXapiVerb
    object: BaseXapiUnnestedObject
    objectType: Literal["SubStatement"]
    result: Optional[BaseXapiResult]
    context: Optional[BaseXapiContext]
    timestamp: Optional[datetime]
    attachments: Optional[List[BaseXapiAttachment]]


BaseXapiObject = Union[
    BaseXapiUnnestedObject,
    BaseXapiSubStatement,
    BaseXapiAgent,
    BaseXapiGroup,
]



"""Utilities for Ralph."""

import datetime
import logging
import operator
from functools import reduce
from importlib import import_module
from typing import List

from pydantic import BaseModel


# Taken from Django utilities
# https://docs.djangoproject.com/en/3.1/_modules/django/utils/module_loading/#import_string
def import_string(dotted_path):
    """Import a dotted module path.

    Returns the attribute/class designated by the last name in the path.
    Raise ImportError if the import failed.
    """
    try:
        module_path, class_name = dotted_path.rsplit(".", 1)
    except ValueError as err:
        raise ImportError(f"{dotted_path} doesn't look like a module path") from err

    module = import_module(module_path)

    try:
        return getattr(module, class_name)
    except AttributeError as err:
        raise ImportError(
            f'Module "{module_path}" does not define a "{class_name}" attribute/class'
        ) from err


def get_backend_type(backends: BaseModel, backend_name: str):
    """Returns the backend type from a backend name."""
    backend_name = backend_name.upper()
    for _, backend_type in backends:
        if hasattr(backend_type, backend_name):
            return backend_type
    return None


def get_backend_instance(backend_type: BaseModel, backend_name: str, options: dict):
    """Returns the instantiated backend instance given backend-name-prefixed options."""
    prefix = f"{backend_name}_"
    # Filter backend-related parameters. Parameter name is supposed to start
    # with the backend name
    names = filter(lambda key: key.startswith(prefix), options.keys())
    options = {name.replace(prefix, ""): options[name] for name in names}
    return getattr(backend_type, backend_name.upper()).get_instance(**options)


def get_root_logger():
    """Get main Ralph logger."""
    ralph_logger = logging.getLogger("ralph")
    ralph_logger.propagate = True

    return ralph_logger


def now():
    """Return the current UTC time in ISO format."""
    return datetime.datetime.now(tz=datetime.timezone.utc).isoformat()


def get_dict_value_from_path(dict_: dict, path: List[str]):
    """Gets a nested dictionary value.

    Args:
        dict_ (dict): dictionnary of values to which the reduction is
            applied
        path (List): array of keys representing the path to the value
    """
    if path is None:
        return None
    try:
        return reduce(operator.getitem, path, dict_)
    except (KeyError, TypeError):
        return None


def set_dict_value_from_path(dict_: dict, path: List[str], value: any):
    """Sets a nested dictionary value.

    Args:
        dict_ (dict): dictionnary where the given value is set
        path (List): array of keys representing the path to the value
        value: value to be set
    """
    for key in path[:-1]:
        dict_ = dict_.setdefault(key, {})
    dict_[path[-1]] = value



"""OVH's LDP storage backend for Ralph."""

import logging

import ovh
import requests

from ralph.conf import settings
from ralph.exceptions import BackendParameterException
from ralph.utils import now

from ..mixins import HistoryMixin
from .base import BaseStorage

ldp_settings = settings.BACKENDS.STORAGE.LDP
logger = logging.getLogger(__name__)


class LDPStorage(HistoryMixin, BaseStorage):
    """OVH's LDP storage backend."""

    # pylint: disable=too-many-arguments

    name = "ldp"

    def __init__(
        self,
        endpoint: str = ldp_settings.ENDPOINT,
        application_key: str = ldp_settings.APPLICATION_KEY,
        application_secret: str = ldp_settings.APPLICATION_SECRET,
        consumer_key: str = ldp_settings.CONSUMER_KEY,
        service_name: str = ldp_settings.SERVICE_NAME,
        stream_id: str = ldp_settings.STREAM_ID,
    ):
        """Instantiates the OVH's LDP client."""
        self._endpoint = endpoint
        self._application_key = application_key
        self._application_secret = application_secret
        self._consumer_key = consumer_key
        self.service_name = service_name
        self.stream_id = stream_id

        self.client = ovh.Client(
            endpoint=self._endpoint,
            application_key=self._application_key,
            application_secret=self._application_secret,
            consumer_key=self._consumer_key,
        )

    @property
    def _archive_endpoint(self):
        if None in (self.service_name, self.stream_id):
            msg = (
                "LDPStorage backend instance requires to set both "
                "service_name and stream_id"
            )
            logger.error(msg)
            raise BackendParameterException(msg)
        return (
            f"/dbaas/logs/{self.service_name}/"
            f"output/graylog/stream/{self.stream_id}/archive"
        )

    def _details(self, name):
        """Returns `name` archive details.

        Expected JSON response looks like:

            {
                "archiveId": "5d49d1b3-a3eb-498c-9039-6a482166f888",
                "createdAt": "2020-06-18T04:38:59.436634+02:00",
                "filename": "2020-06-16.gz",
                "md5": "01585b394be0495e38dbb60b20cb40a9",
                "retrievalDelay": 0,
                "retrievalState": "sealed",
                "sha256": "645d8e21e6fdb8aa7ffc5c[...]9ce612d06df8dcf67cb29a45ca",
                "size": 67906662,
            }
        """
        return self.client.get(f"{self._archive_endpoint}/{name}")

    def url(self, name):
        """Gets archive absolute URL."""
        download_url_endpoint = f"{self._archive_endpoint}/{name}/url"

        response = self.client.post(download_url_endpoint)
        download_url = response.get("url")
        logger.debug("Temporary URL: %s", download_url)

        return download_url

    def list(self, details=False, new=False):
        """Lists archives for a given stream.

        Args:
            details (bool): Get detailed archive information instead of just ids.
            new (bool): Given the history, list only not already fetched archives.
        """
        list_archives_endpoint = self._archive_endpoint
        logger.debug("List archives endpoint: %s", list_archives_endpoint)
        logger.debug("List archives details: %s", str(details))

        archives = self.client.get(list_archives_endpoint)
        logger.debug("Found %d archives", len(archives))

        if new:
            archives = set(archives) - set(self.get_command_history(self.name, "fetch"))
            logger.debug("New archives: %d", len(archives))

        for archive in archives:
            yield self._details(archive) if details else archive

    def read(self, name, chunk_size=4096):
        """Reads the `name` archive file and yields its content."""
        logger.debug("Getting archive: %s", name)

        # Get detailed information about the archive to fetch
        details = self._details(name)

        # Stream response (archive content)
        with requests.get(  # pylint: disable=missing-timeout # nosec
            self.url(name), stream=True
        ) as result:
            result.raise_for_status()
            for chunk in result.iter_content(chunk_size=chunk_size):
                yield chunk

        # Archive is supposed to have been fully fetched, add a new entry to
        # the history.
        self.append_to_history(
            {
                "backend": self.name,
                "command": "fetch",
                "id": name,
                "filename": details.get("filename"),
                "size": details.get("size"),
                "fetched_at": now(),
            }
        )

    def write(self, stream, name, overwrite=False):
        """LDP storage backend is read-only, calling this method will raise an error."""
        msg = "LDP storage backend is read-only, cannot write to %s"
        logger.error(msg, name)
        raise NotImplementedError(msg % name)



"""Base HTTP backend for Ralph."""

import functools
import logging
from abc import ABC, abstractmethod
from enum import Enum, unique
from typing import Iterator, List, Optional, Union

from pydantic import BaseModel, ValidationError

from ralph.exceptions import BackendParameterException

logger = logging.getLogger(__name__)


@unique
class HTTPBackendStatus(Enum):
    """HTTP backend statuses."""

    OK = "ok"
    AWAY = "away"
    ERROR = "error"


class OperationType(Enum):
    """Data backend operation types.

    Attributes:
        INDEX (str): creates a new record with a specific ID.
        CREATE (str): creates a new record without a specific ID.
        DELETE (str): deletes an existing record.
        UPDATE (str): updates or overwrites an existing record.
        APPEND (str): creates or appends data to an existing record.
    """

    INDEX = "index"
    CREATE = "create"
    DELETE = "delete"
    UPDATE = "update"
    APPEND = "append"


def enforce_query_checks(method):
    """Enforces query argument type checking for methods using it."""

    @functools.wraps(method)
    def wrapper(*args, **kwargs):
        """Wrap method execution."""
        query = kwargs.pop("query", None)
        self_ = args[0]

        return method(*args, query=self_.validate_query(query), **kwargs)

    return wrapper


class BaseQuery(BaseModel):
    """Base query model."""

    class Config:
        """Base query model configuration."""

        extra = "forbid"

    query_string: Optional[str]


class BaseHTTP(ABC):
    """Base HTTP backend interface."""

    name = "base"
    query = BaseQuery

    def validate_query(self, query: Union[str, dict, BaseQuery] = None) -> BaseQuery:
        """Validates and transforms the query."""
        if query is None:
            query = self.query()

        if isinstance(query, str):
            query = self.query(query_string=query)

        if isinstance(query, dict):
            try:
                query = self.query(**query)
            except ValidationError as err:
                raise BackendParameterException(
                    "The 'query' argument is expected to be a "
                    f"{self.query.__name__} instance. {err.errors()}"
                ) from err

        if not isinstance(query, self.query):
            raise BackendParameterException(
                "The 'query' argument is expected to be a "
                f"{self.query.__name__} instance."
            )

        logger.debug("Query: %s", str(query))

        return query

    @abstractmethod
    def list(
        self, target: str = None, details: bool = False, new: bool = False
    ) -> Iterator[Union[str, dict]]:
        """Lists containers in the data backend. E.g., collections, files, indexes."""

    @abstractmethod
    def status(self) -> HTTPBackendStatus:
        """Implements HTTP backend check for server status."""

    @abstractmethod
    @enforce_query_checks
    def read(  # pylint: disable=too-many-arguments
        self,
        query: Union[str, BaseQuery] = None,
        target: str = None,
        chunk_size: Union[None, int] = 500,
        raw_output: bool = False,
        ignore_errors: bool = False,
    ) -> Iterator[Union[bytes, dict]]:
        """Yields records read from the HTTP response results."""

    @abstractmethod
    def write(  # pylint: disable=too-many-arguments
        self,
        data: Union[List[bytes], List[dict]],
        target: Union[None, str] = None,
        chunk_size: Union[None, int] = 500,
        ignore_errors: bool = False,
        operation_type: Union[None, OperationType] = None,
    ) -> int:
        """Writes statements into the HTTP server given an input endpoint."""



"""Constants for `cmi5 Profile` xAPI profile."""

# Context extension IDs
CONTEXT_EXTENSION_SESSION_ID = (  # pylint:disable=invalid-name
    "https://w3id.org/xapi/cmi5/context/extensions/sessionid"
)



"""S3 storage backend for Ralph."""

import logging

import boto3
from botocore.exceptions import ClientError, ParamValidationError

from ralph.conf import settings
from ralph.exceptions import BackendException, BackendParameterException
from ralph.utils import now

from ..mixins import HistoryMixin
from .base import BaseStorage

s3_settings = settings.BACKENDS.STORAGE.S3
logger = logging.getLogger(__name__)


class S3Storage(
    HistoryMixin, BaseStorage
):  # pylint: disable=too-many-instance-attributes
    """AWS S3 storage backend."""

    name = "s3"

    # pylint: disable=too-many-arguments

    def __init__(
        self,
        access_key_id: str = s3_settings.ACCESS_KEY_ID,
        secret_access_key: str = s3_settings.SECRET_ACCESS_KEY,
        session_token: str = s3_settings.SESSION_TOKEN,
        default_region: str = s3_settings.DEFAULT_REGION,
        bucket_name: str = s3_settings.BUCKET_NAME,
        endpoint_url: str = s3_settings.ENDPOINT_URL,
    ):
        """Instantiates the AWS S3 client."""
        self.access_key_id = access_key_id
        self.secret_access_key = secret_access_key
        self.session_token = session_token
        self.default_region = default_region
        self.bucket_name = bucket_name
        self.endpoint_url = endpoint_url

        self.client = boto3.client(
            "s3",
            aws_access_key_id=self.access_key_id,
            aws_secret_access_key=self.secret_access_key,
            aws_session_token=self.session_token,
            region_name=self.default_region,
            endpoint_url=self.endpoint_url,
        )

        # Check whether bucket exists and is accessible
        try:
            self.client.head_bucket(Bucket=self.bucket_name)
        except ClientError as err:
            error_msg = err.response["Error"]["Message"]
            msg = "Unable to connect to the requested bucket: %s"
            logger.error(msg, error_msg)
            raise BackendParameterException(msg % error_msg) from err

    def list(self, details=False, new=False):
        """Lists archives in the storage backend."""
        archives_to_skip = set()
        if new:
            archives_to_skip = set(self.get_command_history(self.name, "fetch"))

        try:
            paginator = self.client.get_paginator("list_objects_v2")
            page_iterator = paginator.paginate(Bucket=self.bucket_name)
            for archives in page_iterator:
                if "Contents" not in archives:
                    continue
                for archive in archives["Contents"]:
                    if new and archive["Key"] in archives_to_skip:
                        continue
                    if details:
                        archive["LastModified"] = archive["LastModified"].strftime(
                            "%Y-%m-%d %H:%M:%S"
                        )
                        yield archive
                    else:
                        yield archive["Key"]
        except ClientError as err:
            error_msg = err.response["Error"]["Message"]
            msg = "Failed to list the bucket %s: %s"
            logger.error(msg, self.bucket_name, error_msg)
            raise BackendException(msg % (self.bucket_name, error_msg)) from err

    def url(self, name):
        """Gets `name` file absolute URL."""
        return f"{self.bucket_name}.s3.{self.default_region}.amazonaws.com/{name}"

    def read(self, name, chunk_size: int = 4096):
        """Reads `name` file and yields its content by chunks of a given size."""
        logger.debug("Getting archive: %s", name)

        try:
            obj = self.client.get_object(Bucket=self.bucket_name, Key=name)
        except ClientError as err:
            error_msg = err.response["Error"]["Message"]
            msg = "Failed to download %s: %s"
            logger.error(msg, name, error_msg)
            raise BackendException(msg % (name, error_msg)) from err

        size = 0
        for chunk in obj["Body"].iter_chunks(chunk_size):
            logger.debug("Chunk length %s", len(chunk))
            size += len(chunk)
            yield chunk

        # Archive fetched, add a new entry to the history
        self.append_to_history(
            {
                "backend": self.name,
                "command": "fetch",
                "id": name,
                "size": size,
                "fetched_at": now(),
            }
        )

    def write(self, stream, name, overwrite=False):
        """Writes data from `stream` to the `name` target."""
        if not overwrite and name in list(self.list()):
            msg = "%s already exists and overwrite is not allowed"
            logger.error(msg, name)
            raise FileExistsError(msg % name)

        logger.debug("Creating archive: %s", name)

        try:
            self.client.upload_fileobj(stream, self.bucket_name, name)
        except (ClientError, ParamValidationError) as exc:
            msg = "Failed to upload"
            logger.error(msg)
            raise BackendException(msg) from exc

        # Archive written, add a new entry to the history
        self.append_to_history(
            {
                "backend": self.name,
                "command": "push",
                "id": name,
                "pushed_at": now(),
            }
        )



"""Main module for Ralph's LRS API authentication."""

from ralph.api.auth.basic import get_authenticated_user as get_basic_user
from ralph.api.auth.oidc import get_authenticated_user as get_oidc_user
from ralph.conf import settings

# At startup, select the authentication mode that will be used
get_authenticated_user = (
    get_oidc_user
    if settings.RUNSERVER_AUTH_BACKEND == settings.AuthBackends.OIDC
    else get_basic_user
)



"""Tests for Ralph's configuration loading."""

from importlib import reload
from inspect import signature
from pathlib import PosixPath

import pytest
from pydantic import ValidationError

from ralph import conf
from ralph.conf import CommaSeparatedTuple, Settings, settings
from ralph.utils import import_string


def test_conf_settings_field_value_priority(fs, monkeypatch):
    """Tests that the Settings object field values are defined in the following
    descending order of priority:

        1. Arguments passed to the initializer.
        2. Environment variables.
        3. Dotenv variables (.env)
        4. Default values.
    """
    # pylint: disable=invalid-name

    # 4. Using default value.
    assert str(Settings().AUTH_FILE) == str(settings.APP_DIR / "auth.json")

    # 3. Using dotenv variables (overrides default value).
    fs.create_file(".env", contents="RALPH_AUTH_FILE=/dotenv\n")
    assert str(Settings().AUTH_FILE) == "/dotenv"

    # 2. Using environment variable value (overrides dotenv value).
    monkeypatch.setenv("RALPH_AUTH_FILE", "/environment")
    assert str(Settings().AUTH_FILE) == "/environment"

    # 1. Using argument value (overrides environment value).
    assert str(Settings(AUTH_FILE="/argument").AUTH_FILE) == "/argument"


@pytest.mark.parametrize(
    "value,expected",
    [("foo", ("foo",)), (("foo",), ("foo",)), ("foo,bar,baz", ("foo", "bar", "baz"))],
)
def test_conf_comma_separated_list_with_valid_values(value, expected, monkeypatch):
    """Tests the CommaSeparatedTuple pydantic data type with valid values."""
    assert next(CommaSeparatedTuple.__get_validators__())(value) == expected
    monkeypatch.setenv("RALPH_BACKENDS__DATABASE__ES__HOSTS", "".join(value))
    assert Settings().BACKENDS.DATABASE.ES.HOSTS == expected


@pytest.mark.parametrize("value", [{}, [], None])
def test_conf_comma_separated_list_with_invalid_values(value):
    """Tests the CommaSeparatedTuple pydantic data type with invalid values."""
    with pytest.raises(TypeError, match="Invalid comma separated list"):
        next(CommaSeparatedTuple.__get_validators__())(value)


@pytest.mark.parametrize(
    "ca_certs,verify_certs,expected",
    [
        ("/path", "True", {"ca_certs": PosixPath("/path"), "verify_certs": True}),
        ("/path2", "f", {"ca_certs": PosixPath("/path2"), "verify_certs": False}),
        (None, None, {"ca_certs": None, "verify_certs": None}),
    ],
)
def test_conf_es_client_options_with_valid_values(
    ca_certs, verify_certs, expected, monkeypatch
):
    """Tests the ESClientOptions pydantic data type with valid values."""
    # Using None here as in "not set by user"
    if ca_certs is not None:
        monkeypatch.setenv(
            "RALPH_BACKENDS__DATABASE__ES__CLIENT_OPTIONS__ca_certs", f"{ca_certs}"
        )
    # Using None here as in "not set by user"
    if verify_certs is not None:
        monkeypatch.setenv(
            "RALPH_BACKENDS__DATABASE__ES__CLIENT_OPTIONS__verify_certs",
            f"{verify_certs}",
        )
    assert Settings().BACKENDS.DATABASE.ES.CLIENT_OPTIONS.dict() == expected


@pytest.mark.parametrize(
    "ca_certs,verify_certs",
    [
        ("/path", 3),
        ("/path", None),
    ],
)
def test_conf_es_client_options_with_invalid_values(
    ca_certs, verify_certs, monkeypatch
):
    """Tests the ESClientOptions pydantic data type with invalid values."""
    monkeypatch.setenv(
        "RALPH_BACKENDS__DATABASE__ES__CLIENT_OPTIONS__ca_certs", f"{ca_certs}"
    )
    monkeypatch.setenv(
        "RALPH_BACKENDS__DATABASE__ES__CLIENT_OPTIONS__verify_certs",
        f"{verify_certs}",
    )
    with pytest.raises(ValidationError, match="1 validation error for"):
        Settings().BACKENDS.DATABASE.ES.CLIENT_OPTIONS.dict()


@pytest.mark.parametrize(
    "document_class,tz_aware,expected",
    [
        ("dict", "True", {"document_class": "dict", "tz_aware": True}),
        ("str", "f", {"document_class": "str", "tz_aware": False}),
        (None, None, {"document_class": None, "tz_aware": None}),
    ],
)
def test_conf_mongo_client_options_with_valid_values(
    document_class, tz_aware, expected, monkeypatch
):
    """Tests the MongoClientOptions pydantic data type with valid values."""
    # Using None here as in "not set by user"
    if document_class is not None:
        monkeypatch.setenv(
            "RALPH_BACKENDS__DATABASE__MONGO__CLIENT_OPTIONS__document_class",
            f"{document_class}",
        )
    # Using None here as in "not set by user"
    if tz_aware is not None:
        monkeypatch.setenv(
            "RALPH_BACKENDS__DATABASE__MONGO__CLIENT_OPTIONS__tz_aware",
            f"{tz_aware}",
        )
    assert Settings().BACKENDS.DATABASE.MONGO.CLIENT_OPTIONS.dict() == expected


@pytest.mark.parametrize(
    "document_class,tz_aware",
    [
        ("dict", 3),
        ("str", None),
    ],
)
def test_conf_mongo_client_options_with_invalid_values(
    document_class, tz_aware, monkeypatch
):
    """Tests the MongoClientOptions pydantic data type with invalid values."""
    monkeypatch.setenv(
        "RALPH_BACKENDS__DATABASE__MONGO__CLIENT_OPTIONS__document_class",
        f"{document_class}",
    )
    monkeypatch.setenv(
        "RALPH_BACKENDS__DATABASE__MONGO__CLIENT_OPTIONS__tz_aware",
        f"{tz_aware}",
    )
    with pytest.raises(ValidationError, match="1 validation error for"):
        Settings().BACKENDS.DATABASE.MONGO.CLIENT_OPTIONS.dict()


def test_conf_settings_should_define_all_backends_options():
    """Tests that Settings model defines all backends options."""
    for _, backends in settings.BACKENDS:
        for _, backend in backends:
            # pylint: disable=protected-access
            backend_class = import_string(backend._class_path)
            for parameter in signature(backend_class.__init__).parameters.values():
                if parameter.name == "self":
                    continue
                assert hasattr(backend, parameter.name.upper())


def test_conf_core_settings_should_impact_settings_defaults(monkeypatch):
    """Tests that core settings update application settings values."""
    monkeypatch.setenv("RALPH_APP_DIR", "/foo")
    monkeypatch.setenv("RALPH_LOCALE_ENCODING", "ascii")
    reload(conf)

    # Configuration.
    assert conf.Settings.Config.env_file_encoding == "ascii"

    # Properties.
    assert str(conf.settings.APP_DIR) == "/foo"
    assert conf.settings.LOCALE_ENCODING == "ascii"

    # Defaults.
    assert str(conf.settings.AUTH_FILE) == "/foo/auth.json"
    assert conf.settings.BACKENDS.STORAGE.FS.PATH == "/foo/archives"



"""Base xAPI `Object` definitions (1)."""

from typing import Dict, List, Optional, Union

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from uuid import UUID

from pydantic import AnyUrl, StrictStr, constr, validator

from ..config import BaseModelWithConfig
from .common import IRI, LanguageMap


class BaseXapiActivityDefinition(BaseModelWithConfig):
    """Pydantic model for `Activity` type `definition` property.

    Attributes:
        name (LanguageMap): Consists of the human readable/visual name of the Activity.
        description (LanguageMap): Consists of a description of the Activity.
        type (IRI): Consists of the type of the Activity.
        moreInfo (URL): Consists of an URL to a document about the Activity.
        extensions (dict): Consists of a dictionary of other properties as needed.
    """

    name: Optional[LanguageMap]
    description: Optional[LanguageMap]
    type: Optional[IRI]
    moreInfo: Optional[AnyUrl]
    extensions: Optional[Dict[IRI, Union[str, int, bool, list, dict, None]]]


class BaseXapiInteractionComponent(BaseModelWithConfig):
    """Pydantic model for an interaction component.

    Attributes:
        id (str): Consists of an identifier of the interaction component.
        description (LanguageMap): Consists of the description of the interaction.
    """

    id: constr(regex=r"^[^\s]+$")  # noqa:F722
    description: Optional[LanguageMap]


class BaseXapiActivityInteractionDefinition(BaseXapiActivityDefinition):
    """Pydantic model for `Activity` type `definition` property.

    It is defined for field with interaction properties.

    Attributes:
        interactionType (str): Consists of the type of the interaction.
        correctResponsesPattern (list): Consists of a pattern for the correct response.
        choices (list): Consists of a list of selectable choices.
        scale (list): Consists of a list of the options on the `likert` scale.
        source (list): Consists of a list of sources to be matched.
        target (list): Consists of a list of targets to be matched.
        steps (list): Consists of a list of the elements making up the interaction.
    """

    interactionType: Literal[
        "true-false",
        "choice",
        "fill-in",
        "long-fill-in",
        "matching",
        "performance",
        "sequencing",
        "likert",
        "numeric",
        "other",
    ]
    correctResponsesPattern: Optional[List[StrictStr]]
    choices: Optional[List[BaseXapiInteractionComponent]]
    scale: Optional[List[BaseXapiInteractionComponent]]
    source: Optional[List[BaseXapiInteractionComponent]]
    target: Optional[List[BaseXapiInteractionComponent]]
    steps: Optional[List[BaseXapiInteractionComponent]]

    @validator("choices", "scale", "source", "target", "steps")
    @classmethod
    def check_unique_ids(cls, value):
        """Checks the uniqueness of interaction components IDs."""
        if len(value) != len({x.id for x in value}):
            raise ValueError("Duplicate InteractionComponents are not valid")


class BaseXapiActivity(BaseModelWithConfig):
    """Pydantic model for `Activity` type property.

    Attributes:
        id (IRI): Consists of an identifier for a single unique Activity.
        objectType (str): Consists of the value `Activity`.
        definition (dict): See BaseXapiActivityDefinition and
            BaseXapiActivityInteractionDefinition.
    """

    id: IRI
    objectType: Optional[Literal["Activity"]]
    definition: Optional[
        Union[
            BaseXapiActivityDefinition,
            BaseXapiActivityInteractionDefinition,
        ]
    ]


class BaseXapiStatementRef(BaseModelWithConfig):
    """Pydantic model for `StatementRef` type property.

    Attributes:
        objectType (str): Consists of the value `StatementRef`.
        id (UUID): Consists of the UUID of the referenced statement.
    """

    id: UUID
    objectType: Literal["StatementRef"]


BaseXapiUnnestedObject = Union[BaseXapiActivity, BaseXapiStatementRef]



"""Base database backend for Ralph."""

import functools
import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime
from enum import Enum, unique
from typing import BinaryIO, List, Literal, Optional, TextIO, Union
from uuid import UUID

from pydantic import BaseModel

from ralph.exceptions import BackendParameterException

logger = logging.getLogger(__name__)


class BaseQuery(BaseModel):
    """Base query model."""

    class Config:
        """Base query model configuration."""

        extra = "forbid"


@dataclass
class StatementQueryResult:
    """Represents a common interface for results of an LRS statements query."""

    statements: List[dict]
    pit_id: str
    search_after: str


@unique
class DatabaseStatus(Enum):
    """Database statuses."""

    OK = "ok"
    AWAY = "away"
    ERROR = "error"


@dataclass
class StatementParameters:
    """Represents a dictionary of possible LRS query parameters."""

    # pylint: disable=too-many-instance-attributes

    statementId: Optional[str] = None  # pylint: disable=invalid-name
    voidedStatementId: Optional[str] = None  # pylint: disable=invalid-name
    agent__mbox: Optional[str] = None
    agent__mbox_sha1sum: Optional[str] = None
    agent__openid: Optional[str] = None
    agent__account__name: Optional[str] = None
    agent__account__home_page: Optional[str] = None
    verb: Optional[str] = None
    activity: Optional[str] = None
    registration: Optional[UUID] = None
    related_activities: Optional[bool] = False
    related_agents: Optional[bool] = False
    since: Optional[datetime] = None
    until: Optional[datetime] = None
    limit: Optional[int] = None
    format: Optional[Literal["ids", "exact", "canonical"]] = "exact"
    attachments: Optional[bool] = False
    ascending: Optional[bool] = False
    search_after: Optional[str] = None
    pit_id: Optional[str] = None

    def __post_init__(self):
        """Perform additional conformity verifications on parameters."""
        # Check that both `homePage` and `name` are provided if `account` is being used
        if (self.agent__account__name is not None) != (
            self.agent__account__home_page is not None
        ):
            raise BackendParameterException(
                "Invalid agent parameters: home_page and name are both required"
            )

        # Check that no more than one Inverse Functional Identifier is provided
        if (
            sum(
                x is not None
                for x in [
                    self.agent__mbox,
                    self.agent__mbox_sha1sum,
                    self.agent__openid,
                    self.agent__account__name,
                ]
            )
            > 1
        ):
            raise BackendParameterException(
                "Invalid agent parameters: Only one identifier can be used"
            )


def enforce_query_checks(method):
    """Enforce query argument type checking for methods using it."""

    @functools.wraps(method)
    def wrapper(*args, **kwargs):
        """Wrap method execution."""
        query = kwargs.pop("query", None)
        self_ = args[0]

        return method(*args, query=self_.validate_query(query), **kwargs)

    return wrapper


class BaseDatabase(ABC):
    """Base database backend interface."""

    name = "base"
    query_model = BaseQuery

    def validate_query(self, query: BaseQuery = None):
        """Validate database query."""
        if query is None:
            query = self.query_model()

        if not isinstance(query, self.query_model):
            raise BackendParameterException(
                "'query' argument is expected to be a "
                f"{self.query_model().__class__.__name__} instance."
            )

        logger.debug("Query: %s", str(query))

        return query

    @abstractmethod
    def status(self) -> DatabaseStatus:
        """Implements database checks (e.g. connection, cluster status)."""

    @abstractmethod
    @enforce_query_checks
    def get(self, query: BaseQuery = None, chunk_size: int = 10):
        """Yields `chunk_size` records read from the database query results."""

    @abstractmethod
    def put(
        self,
        stream: Union[BinaryIO, TextIO],
        chunk_size: int = 10,
        ignore_errors: bool = False,
    ) -> int:
        """Writes `chunk_size` records from the `stream` to the database.

        Returns:
            int: The count of successfully written records.
        """

    @abstractmethod
    def query_statements(self, params: StatementParameters) -> StatementQueryResult:
        """Returns the statements query payload using xAPI parameters."""

    @abstractmethod
    def query_statements_by_ids(self, ids: List[str]) -> List:
        """Returns the list of matching statement IDs from the database."""
